--------------------------------------------------------------------------------
-- Random Tour Lua Script
-- By Scepheo
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Scenario configuration
--------------------------------------------------------------------------------

-- How big the checkpoint should be in relation to the node
local nodeSizeMultiplier = 1.5

-- Distance after which a new checkpoint should be placed
local nodeMaxDistance = 50

-- Distance after which the first checkpoint should be placed
local nodeInitialDistance = 20

-- How many nodes (not checkpoints!) the path needs to contain
local minPathLength = 15

--------------------------------------------------------------------------------
-- Scenario code
--------------------------------------------------------------------------------
local M = {}

local graphpath = require('graphpath')

local loaded = false
local running = false

local function getRandomKey(items)
	-- Count the number of items in the table
	local itemCount = 0
	
	for key, value in pairs(items) do
		itemCount = itemCount + 1
	end
	
	-- Select a random number n
	local selectedIndex = math.random(1, itemCount)
	
	-- Get the nth node
	local index = 1
	local selectedKey = ""
	
	for key, value in pairs(items) do
		if index == selectedIndex then
			selectedKey = key
			break
		end
		
		index = index + 1
	end
	
	return selectedKey
end

local function getRandomValue(items)
	return items[getRandomKey(items)]
end

local function addWaypointFor(scenario, nodeName)
	local node = map.map.nodes[nodeName]
	local wpName = 'wp_' .. nodeName
	local position = node.pos.x .. ' ' .. node.pos.y .. ' ' .. node.pos.z

	TorqueScript.eval(
	[[
		if (isObject(]] .. wpName .. [[))
		{
			]] .. wpName .. [[.delete();
		}
		
		new BeamNGWaypoint(]] .. wpName .. [[)
		{
			drawDebug = "0";
			position = "]] .. position .. [[";
			rotation = "1 0 0 0";
			scale = "1 1 1";
			canSave = "1";
			canSaveDynamicFields = "1";
		};
		
		ScenarioObjectsGroup.add(]] .. wpName .. [[);
	]])
	
	table.insert(scenario.lapConfig, wpName)
	scenario.nodes[wpName] = 
	{
		pos = vec3(node.pos),
		radius = node.radius * nodeSizeMultiplier
	}
end

local function loadPlayer(position, direction)
	
	-- Set the player at the starting location
	local tsPosition = position.x .. ' ' .. position.y .. ' ' .. position.z
	
	local angle = math.atan2(direction.x, direction.y)
	angle = math.deg(angle) + 180
	local tsRotation = '0 0 1 ' .. angle
	
	local vehicleList = vehicles.getVehicleList().list
	
	local vehicle = {}
	local valid = false
	
	while not valid do
		vehicle = getRandomValue(vehicleList)
		valid = vehicle.Type == "Car" or vehicle.Type == "Truck"
	end
	
	local jbeam = vehicle.key
	local configDisplay = vehicle.Configuration or '(Default configuration)'
	
	log('I', 'random_tour.loadPlayer',
		string.format('Selected %s %s', jbeam, configDisplay))
	
	local configuration = ''
	if (vehicle.ConfigName) then
		configuration = string.format('vehicles/%s/%s.pc', jbeam, vehicle.ConfigName)
	end
	
	local color
	
	if vehicle.colors == nil then
		color = "0 0 0 1"
	else
		color = getRandomValue(vehicle.colors)
	end
	
	TorqueScript.eval(
	[[
		scenario_player0.setTransform(matrixCreate("]] .. tsPosition .. [[", "]] .. tsRotation .. [["));
		scenario_player0.JBeam = "]] .. jbeam .. [[";
		scenario_player0.partConfig = "]] .. configuration .. [[";
		scenario_player0.color = "]] .. color .. [[";
		scenario_player0.requestReload();
	]])
end

local function getRandomPath()

	-- Build gp
	local gp = graphpath.newGraphpath()
	
    for k, n in pairs(map.map.nodes) do
        n.pos = vec3(n.pos)
        gp:setPointPosition(k, n.pos)
        n.drivability = 1
    end
	
    for nid, n in pairs(map.map.nodes) do
        for lid, dif in pairs(n.links) do
            local totaldif = map.map.nodes[nid].pos:distance(map.map.nodes[lid].pos) * (1 / (dif + 1e-30))
            gp:bidiEdge(nid, lid, totaldif)
        end
    end
	
	-- Seed RNG
	math.randomseed(os.time())
	
	-- Select a path that's possible and longer than the minimum length
	local path = nil
	
	while path == nil or #path < minPathLength do
		startNodeName = getRandomKey(map.map.nodes)
		endNodeName = startNodeName
		
		while endNodeName == startNodeName do
			endNodeName = getRandomKey(map.map.nodes)
		end
		
		path = gp:getPath(startNodeName, endNodeName)
	end
	
	return path
end

local function onLoad(scenario)
	
	local path = getRandomPath()
	
	log('I', 'random_tour.onLoad', 'Got random path:')
	
	for i, nodeName in ipairs(path) do
		log('I', 'random_tour.onLoad', string.format('%003d. %s', i, nodeName))
	end
	
	scenario.lapConfig = {}
	
	local startPosition = map.map.nodes[path[1]].pos
	
	local previousPosition = startPosition
	
	local startDirection = nil
	
	local currentDistance = nodeMaxDistance - nodeInitialDistance
	
	-- Create a waypoint for some nodes
	for i = 1, #path - 1 do
		local nodeName = path[i]
		local node = map.map.nodes[nodeName]
		
		local distance = (node.pos - previousPosition):length()
		currentDistance = currentDistance + distance
				
		if currentDistance > nodeMaxDistance then
			
			if startDirection == nil then
				startDirection = node.pos - startPosition
			end
		
			addWaypointFor(scenario, nodeName)
			
			currentDistance = 0
		end
		
		previousPosition = node.pos
	end
	
	addWaypointFor(scenario, path[#path])
	
	if startDirection == nil then
		startDirection = map.map.nodes[path[#path]].pos - startPosition
	end
	
	loadPlayer(startPosition, startDirection)
end

local function onScenarioChange(scenario)
	if scenario and scenario.state == 'pre-start' then
		log('I', 'random_tour.onScenarioChange', 'Scenario started, starting set-up')
		onLoad(scenario)
		log('I', 'random_tour.onScenarioChange', 'Scenario set-up complete')
	end
end

local function onPhysicsEngineEvent(eventType)
	-- Freeze whenever a vehicle is spawned
	if eventType == 'spawn' then
		log('I', 'random_tour.onPhysicsEngineEvent', 'Vehicle spawned, freezing scenario_player0')
		scenetree.scenario_player0:queueLuaCommand('scenario.freeze(1)')
	end
end

M.onPhysicsEngineEvent = onPhysicsEngineEvent
M.onScenarioChange = onScenarioChange

return M